#include <algorithm>
#include <msp/fs/utils.h>
#include "builder.h"
#include "component.h"
#include "objectfile.h"
#include "sourcefile.h"
#include "sourcepackage.h"

using namespace std;
using namespace Msp;

ObjectFile::ObjectFile(Builder &b, const Component &c, SourceFile &s):
	FileTarget(b, c.get_package(), generate_target_path(c, s.get_path())),
	source(s),
	used_in_shlib(false)
{
	component = &c;
	add_dependency(source);
}

FS::Path ObjectFile::generate_target_path(const Component &comp, const FS::Path &src)
{
	const SourcePackage &pkg = comp.get_package();
	FS::Path temp_dir = pkg.get_temp_directory()/comp.get_name();
	FS::Path rel_src;
	if(FS::descendant_depth(src, temp_dir)>=0)
		rel_src = FS::relative(src, temp_dir);
	else
		rel_src = FS::relative(src, pkg.get_source_directory());
	string fn;
	for(FS::Path::Iterator i=rel_src.begin(); i!=rel_src.end(); ++i)
	{
		if(!fn.empty())
			fn += '_';
		if(*i!=".")
			fn += *i;
	}
	return temp_dir/(FS::basepart(fn)+".o");
}

void ObjectFile::set_used_in_shared_library(bool u)
{
	used_in_shlib = u;
}

void ObjectFile::collect_build_info(BuildInfo &binfo) const
{
	Target::collect_build_info(binfo);
	binfo.update_from(component->get_build_info_for_path(source.get_path()));
}

void ObjectFile::find_dependencies()
{
	for(Dependencies::iterator i=depends.begin(); i!=depends.end(); ++i)
	{
		(*i)->prepare();
		find_dependencies(dynamic_cast<FileTarget *>(*i));
	}
}

void ObjectFile::find_dependencies(FileTarget *tgt)
{
	FileTarget *rtgt = dynamic_cast<FileTarget *>(tgt->get_real_target());
	const Dependencies &tdeps = rtgt->get_transitive_dependencies();
	Dependencies deps_to_add;
	if(rtgt==tgt)
	{
		/* We are using the target from its original location, so dependencies
		apply directly */
		deps_to_add = tdeps;
	}
	else
	{
		/* The target has been displaced by installing it.  Displace any
		dependencies that come from the same package as well. */
		const SourcePackage *tpkg = rtgt->get_package();
		for(Dependencies::const_iterator i=tdeps.begin(); i!=tdeps.end(); ++i)
		{
			FileTarget *file = dynamic_cast<FileTarget *>(*i);
			if(file && file->get_package()==tpkg && FS::descendant_depth(file->get_path(), tpkg->get_source_directory())>=0)
			{
				FS::Path displaced = tgt->get_path()/FS::relative(file->get_path(), rtgt->get_path());
				if(Target *ddep = builder.get_vfs().get_target(displaced))
					deps_to_add.push_back(ddep);
				else
				{
					const Component *tcomp = file->get_component();
					const Component::OverlayList &overlays = tcomp->get_overlays();
					string last_dir = FS::basename(FS::dirname(displaced));
					for(Component::OverlayList::const_iterator j=overlays.begin(); j!=overlays.end(); ++j)
						if(last_dir==*j)
						{
							displaced = displaced.subpath(0, displaced.size()-2)/FS::basename(file->get_path());
							if((ddep = builder.get_vfs().get_target(displaced)))
								deps_to_add.push_back(ddep);
						}
				}
			}
			else
				deps_to_add.push_back(*i);
		}
	}

	for(Dependencies::const_iterator i=deps_to_add.begin(); i!=deps_to_add.end(); ++i)
		if(find(depends.begin(), depends.end(), *i)==depends.end())
		{
			add_dependency(**i);
			if((*i)->get_real_target()->is_buildable())
				(*i)->signal_modified.connect(sigc::mem_fun(this, static_cast<void (ObjectFile::*)()>(&ObjectFile::find_dependencies)));
		}
}
